
# This file was generated by 'versioneer.py' (0.18) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2020-10-20T11:38:21-0400",
 "dirty": false,
 "error": null,
 "full-revisionid": "843d3d3d8f01fc4d8ab2eaf0f42060abfa32e405",
 "version": "0.12.6"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)
